import java.awt.*;
import java.awt.event.*;
import java.lang.*;
import javax.swing.*;
import java.util.*;
import javax.swing.JSpinner.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

public class Gui extends Panel implements Runnable, ChangeListener
{
    private Thread mTimer;
    private Label[][] mLabels;
    private JSpinner[] mNumbers;
    private String savedData;
    private Xport mXport;
    
    public Gui(Xport xport)
    {
        setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.insets =  new Insets(5,5,5,5);
        setBackground(java.awt.Color.lightGray);
        setSize(328, 446);
        
        mXport = xport;
        InitializeLabels(c);
        
        mTimer = new Thread(this);
        mTimer.start();
    }
    
    private void InitializeLabels(GridBagConstraints c)
    {
        mLabels = new Label[8][6];
        mNumbers = new JSpinner[8];
        for(int x = 0; x < 8; x++)
        {
            //Go through and Initialize all the labels
            for(int y = 0; y < 6; y++)
            {
                mLabels[x][y] = new Label();
                add(mLabels[x][y], c);
            }
            
            // we make all of the spinners
            mNumbers[x] = new JSpinner(new SpinnerNumberModel(0, 0, 255, 1));
            mNumbers[x].addChangeListener(this);
            
            c.gridx = 6; c.gridy = 2*x + 1; c.gridwidth = 1; c.gridheight = 1;
            c.weightx = 0.0; c.weighty = 0.0; c.anchor = GridBagConstraints.WEST;
            add(mNumbers[x], c);

            // We add all of our labels to the GUI
            addLabel("Sensor " + (x + 1) + ":", mLabels[x][0], 0, 2*x, c);
            addLabel("Sensor Output:", mLabels[x][1], 1, 2*x+1, c);
            addLabel("N/A"           , mLabels[x][2], 2, 2*x+1, c);
            addLabel("Status:"       , mLabels[x][3], 3, 2*x+1, c);
            addLabel("N/A"           , mLabels[x][4], 4, 2*x+1, c);
            addLabel("Threshold:"    , mLabels[x][5], 5, 2*x+1, c);
        }
    }

    private void addLabel(String text, Label label, int x, int y, GridBagConstraints c)
    {
        mLabels[x][5].setText(text);
        c.gridx = x; c.gridy = y; c.gridwidth = 1; c.gridheight = 1;
        c.weightx = 0.0; c.weighty = 0.0; c.anchor = GridBagConstraints.WEST;
        add(label, c);
    }
    
    public void run() 
    {
        Thread me = Thread.currentThread();
        while (mTimer == me) 
        {
            try 
            {
                // we run this thread once every 1 second
                Thread.currentThread().sleep(1000);
                
                // if we havn't connected then just continue
                if(mXport == null || mXport.connected() == false)
                    continue;
                
                // we get input from the xport and process the
                // data if valid. 
                UpdateInput();
            }
            catch (InterruptedException e) { }
        }
    }
    
    // Callback for when a spinner is changed. 
    public void stateChanged(ChangeEvent e)
    {
        // if not connected then don't bother sending data
        if(mXport == null || mXport.connected() == false)
            return;
        
        JSpinner spinner = (JSpinner)e.getSource(); // get the spinner that's being changed
        int spinnerNum = getSpinnerNum(spinner);    // get the number of the spinnger
        
        assert(spinnerNum != -1);       // make sure that we found a spinner
        
        String packet = new String(""); // make a new string
        packet += (char)127;            // put the start character
        packet += spinnerNum;           // the sensor number
        packet += spinner.getValue();   // put the new threshold
        packet += (char)126;            // put the end character

        mXport.send(packet); // send out new threshold.
    }
    
    // returns the index of a spinner if we added it to the GUI
    // otherwise returns -1.
    public int getSpinnerNum(JSpinner spinner)
    {
        // we go through all of our spinners
        for(int x = 0; x < 8; x++)
        {
            // if we find our spinner return it's index
            if(spinner == mNumbers[x])
                return x;
        }
        
        // if we get here it means that we did not find it's index
        return -1;
    }
    
    public void UpdateInput()
    {
        // we receive new data from the xport
        byte[] newData = mXport.receive();
        
        //if we have no new data we return or it's an invalid input
        if(newData == null || isValidInput(newData) == false)
            return;
        
        // update the GUI with the new info
        updateOutput(newData);
    }
    
    private void updateOutput(byte[] data)
    {   
        // we process the data from what is pulled 
        // out of the Xport.
        int[][] formattedData = formatData(data); 

        // we go through all of the sensors and update their values
        for(int x = 0; x < 8; x++)
        {
         Integer sensorData = formattedData[x][0];
         Integer threshHold = formattedData[x][1];
         int sensorOutput = formattedData[x][2];
         
         mLabels[x][2].setText(sensorData.toString());
         mLabels[x][4].setText((sensorOutput == 1) ? "On" : "Off");
         mNumbers[x].setValue(threshHold);
        }
    }
    
    private int[][] formatData(byte[] rawData)
    {
        int[][] finalData = new int[8][3];
        int index = 0;

        //NOTE: we need to pull the data out of the format that it is in. 
        //it takes 3 bytes to encode one number, to decode it we pull 
        //out the 3 bytes and regenerate the number, and put it into 
        //the final data array

        for(int x = 0; x < 8; x++)
        for(int y = 0; y < 3; y++)
        {
            index = x * 9 + y * 3 + 1;
            finalData[x][y] = byteToInt(rawData[index])*100 
                            + byteToInt(rawData[index+1])*10 
                            + byteToInt(rawData[index+2]);
        }
        
        return finalData;
    }
    
    private int byteToInt(byte data)
    {
        // if it's a space we return 0 as it is just padding
        if(data == 32)
            return 0;
        
        //otherwise we return the int value of the byte.
        return data - 48;
    }
    
    private boolean isValidInput(byte[] data)
    {
        // if we do not have a complete string then it's not valid
        if(data[0] == 127 && data[data.length-1] == 126)
            return true;
        
        // NOTE: this is a very conservative method of checking to see if the
        // input is valid. In testing I saw a decent amount of the data being
        // sent in two packets instead of one. In this case we would lose 
        // valid data. To resolve this we need a smarter way of handling 
        // packet fragmentation. This basically invovles keeping some sort
        // of structure that buffers the data till it has a complete tranmission.
        
        return false;
    }
}